/*-
 * Copyright (c) 1998 Eivind Eklund
 * Copyright (c) 1996, 1997 DiMaga Studios & Innerloop Technologies
 * All rights reserved.
 *
 * This code was originally written for DiMaga Studios by Eivind
 * Eklund.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Neither the name of Eivind Eklund nor the names of DiMaga
 *    Studios or Innerloop Technologies may be used to endorse or
 *    promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#ifndef BSEARCH_H
#define BSEARCH_H

struct Something;
void
BinSearch(size_t* pResult,
          struct Something aSorted[],
          size_t nElements,
          struct Something* pMatch,
          int ElementName);
/* This will ALWAYS return a result within the range; the closest
   element index below the given key. */

#define BinSearch(pResult, aSorted, nElements, vMatch, ElementName)  \
do {                                                                 \
  size_t _iMin = 0;             /* Minimum element index */          \
  size_t _iMax = (nElements);   /* Maximum element index */          \
  assert(_iMax);                                                     \
  while( _iMin+1 < _iMax ) {                                         \
    if( (aSorted)[(_iMin+_iMax)/2].ElementName < (vMatch) )          \
      _iMin = (_iMin+_iMax)/2;                                       \
    else if( (aSorted)[(_iMin+_iMax)/2].ElementName > (vMatch) )     \
      _iMax = (_iMin+_iMax)/2;                                       \
    else                                                             \
      break;                                                         \
  }                                                                  \
  *(pResult) = (_iMin+_iMax)/2;                                      \
} /*lint -save -e717 */ while(0) /*lint -restore */

;

#define ASSERT_SORTED(array, nitems, id)              \
do {                                                  \
   size_t _NumItems = (nitems)-1;                     \
   size_t _i;                                         \
   for(_i=0; _i<_NumItems; _i++) {                    \
      assert( (array)[_i].id <= (array)[_i+1].id );   \
   }                                                  \
} /*lint -save -e717 */ while(0) /*lint -restore */

#endif /* !defined(BSEARCH_H) */
