/* -*- mode: c; tab-width: 3; c-basic-indent: 3; -*-
 * Copyright (c) 1998 Eivind Eklund
 * Copyright (c) 1996, 1997 DiMaga Studios & Innerloop Technologies
 * All rights reserved.
 *
 * This code was originally written for DiMaga Studios by Eivind
 * Eklund.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Neither the name of Eivind Eklund nor the names of DiMaga
 *    Studios or Innerloop Technologies may be used to endorse or
 *    promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "./node.h"

/*lint -library */
/*lint -e720 Boolean test of assignment */

#undef List_AddTail
#undef List_AddHead
#undef List_AddTailList
#undef List_AddHeadList
#undef List_Length
#undef List_Verify
#undef List_RemHead
#undef Node_Remove
#undef Node_InsertBefore
#undef Node_InsertAfter
#undef Node_GetList
#undef Node_SwapOut


void
List_AddTail(struct List* pList /* list to add to */, struct Node* pNode /* node to add */) {
	/* Unit test available & passed. Tue Jul  2 11:29:41 1996 */
	LIST_ASSERT( pList );
	NODE_ASSERT( pNode );
	assert( !pNode->pPrev && !pNode->pSucc );
	pNode->pPrev = pList->pTail;
	pNode->pSucc = (struct Node*)&pList->MIDDLE; /*lint !e740 Unusual pointer */
	pList->pTail->pSucc = pNode;
	pList->pTail = pNode;
}

void
List_AddHead(struct List* pList /* list to add to */, struct Node* pNode /* node to add */) {
	/* Unit test available & passed. Tue Jul  2 13:59:18 1996 */
	LIST_ASSERT( pList );
	NODE_ASSERT( pNode );
	assert( !pNode->pPrev && !pNode->pSucc );
	pNode->pSucc = pList->pHead;
	pNode->pPrev = (struct Node*)pList; /*lint !e740 Unusual pointer cast */
	pList->pHead->pPrev = pNode;
	pList->pHead = pNode;
}

struct Node*
List_RemHead(struct List* pList) {
	/* Unit test available & passed. Tue Jul  2 11:29:56 1996 */
	struct Node* pNode;
	LIST_ASSERT(pList);
	if(List_isEmpty(pList))
		return NULL;
	pNode = pList->pHead;
	pList->pHead = pNode->pSucc;
	pList->pHead->pPrev = (struct Node*)pList; /*lint !e740 */
	pNode->pSucc = pNode->pPrev = NULL;
	return pNode;
}

void
List_AddTailList(struct List* pList, struct List* pNew) {
	/* Unit test available & passed. Tue Jul  2 13:59:38 1996 */
	LIST_ASSERT(pList);
	LIST_ASSERT(pNew);
	if(List_isEmpty(pNew))
			return;
	pNew->pTail->pSucc = pList->pTail->pSucc;
	pList->pTail->pSucc = pNew->pHead;
	pNew->pHead->pPrev = pList->pTail;
	pList->pTail = pNew->pTail;
	List_Init(pNew);
}

void
List_AddHeadList(struct List* pList, struct List* pNew) {
	/* Unit test available & passed. Tue Jul  2 17:13:25 1996 */
	LIST_ASSERT(pList);
	LIST_ASSERT(pNew);
	if( List_isEmpty(pNew) )
		return;
	pNew-> pHead->pPrev = pList->pHead->pPrev;
	pList->pHead->pPrev = pNew-> pTail;
	pNew-> pTail->pSucc = pList->pHead;
	pList->pHead = pNew->pHead;
	List_Init(pNew);
}

void
Node_Remove(struct Node* pNode) {
	/* Unit test available & passed. Tue Jul  2 17:14:10 1996 */
	NODE_ASSERT( pNode );
	assert( pNode->pSucc && pNode->pPrev );
	pNode->pPrev->pSucc = pNode->pSucc;
	pNode->pSucc->pPrev = pNode->pPrev;
	pNode->pSucc = pNode->pPrev = NULL;
}

void
Node_InsertBefore(struct Node* pNode, struct Node* pNew) {
	/* Unit test available & passed. Tue Jul  2 17:23:16 1996 */
	NODE_ASSERT( pNode );
	NODE_ASSERT( pNew );
	assert( !pNew->pSucc && !pNew->pPrev );
	assert( pNode->pSucc && pNode->pPrev );
	pNode->pPrev->pSucc = pNew;
	pNew->pPrev = pNode->pPrev;
	pNode->pPrev = pNew;
	pNew->pSucc = pNode;
}

void
Node_InsertAfter(struct Node* pNode, struct Node* pNew) {
	/* Unit test available & passed. Tue Jul  2 17:23:16 1996 */
	NODE_ASSERT( pNode );
	NODE_ASSERT( pNew );
	assert( !pNew->pSucc && !pNew->pPrev );
	assert( pNode->pSucc && pNode->pPrev );
	pNode->pSucc->pPrev = pNew;
	pNew->pSucc = pNode->pSucc;
	pNode->pSucc = pNew;
	pNew->pPrev = pNode;
}

/* check whether a list is valid. */
/* This function is used for validating a list - it will check that */
/* the pointer is non-NULL, and that general list integrity is OK,  */
/* starting with checking that the head is initialized and going on */
/* to check that each node is chained correctly to the one before,  */
/* and that the list is chained to the head in both ends. */
const char*											/* NULL if OK, description of error if the list is corrupt. */
List_Verify(const struct List* pList /* The list to verify */) {
	/* Implictly working through the rest of the unit test working. */
	struct Node* pNode1;					/* Temp value during check */
	struct Node* pNode2;					/* Temp value during check */
	if( !pList )
		return "List_Verify: NULL List";
	if( !pList->pHead || pList->MIDDLE || !pList->pTail )
		return "List_Verify: Badly initialized list - NULL/non-NULL in wrong places in the header";
	if( pList->pHead->pPrev != (void*)pList )
		return "List_Verify: List is disconnected from head.";
	if( pList->pTail->pSucc != (void*)&pList->MIDDLE )
		return "List_Verify: List is disconnected from tail.";
	for(pNode1=pList->pHead; (pNode2=pNode1->pSucc) && pNode2->pSucc; pNode1=pNode2) { /*lint !e720 */
		if( !pNode2->pPrev )
			return "List_Verify: Node has a NULL pPrev somewhere.";
		if( pNode2->pPrev != pNode1 )
			return "List_Verify: List is broken.";
	}
	if( !List_isEmpty(pList) && pNode2 != (void*)&pList->MIDDLE )
		return "List_Verify: List has a NULL pointer or is connected to another lists tail.";
	return NULL;
}

/* get the list header of a list you have a node in. */
/* Node_GetList will trace a list you supply a node from, and get the */
/* header of that list.  The node *must* be a member of a list, or */
/* the function will assert. */

struct List*										/* list the supplied node is a member of. */
Node_GetList(
struct Node* pNode /* node to start tracing from */
) {
	/* Unit test available & passed. Tue Jul  9 13:38:55 1996 */

	NODE_ASSERT( pNode );
	assert( pNode->pSucc && pNode->pPrev );				/* Is a member of a list */

	while( pNode->pPrev )
		pNode = pNode->pPrev;

	return (struct List*)pNode;		/*lint !e740 Unusual pointer cast */
}

/* swap a node out of a list. */
/* Node_SwapOut replaces a node 'in place' in a list.  This can be
	convenient for doing changes to a list during traversal of the same
	list.

	The pointer to the node swapped out is stored in the pointer to the
	node swapped in, making this call symmetric - calling it twice will
	restore the status quo. */

void
Node_SwapOut(
struct Node* pNode,				  /* Node to swap out.  Must already be
											  a member of a list. */
struct Node** ppExternal		  /* Pointer to pointer to external
											  node.  The node this point to will
											  be substituted for pNode in the
											  list pNode is in, while a pointer
											  to pNode will be fed in here. */
) {
	NODE_ASSERT( pNode );
	assert( pNode->pSucc && pNode->pPrev );				/* Is a member of a list */
	assert( ppExternal );
	assert( (*ppExternal)->pSucc == NULL && (*ppExternal)->pPrev == NULL );

	pNode->pPrev->pSucc = *ppExternal;
	pNode->pSucc->pPrev = *ppExternal;
	(*ppExternal)->pSucc = pNode->pSucc;
	(*ppExternal)->pPrev = pNode->pPrev;
	*ppExternal = pNode;
	pNode->pSucc = pNode->pPrev = NULL;
	return;
}


/* Find the length of list.
	Reasonably obvious function, isn't it? */
unsigned int						  /* Number of nodes in list */
List_Length(
	const struct List* pList	  /* List to find length of */
) {
	struct Node* pNode,* pNodeNext;
	size_t nEntries;

	LIST_ASSERT( pList );
	nEntries = 0;
	LIST_WALK_FORWARD(pList, pNode, pNodeNext) {
		nEntries++;
	}
	return nEntries;
}
