/* -*- mode: c; tab-width: 3; c-basic-indent: 3; -*-
 * Copyright (c) 1998 Eivind Eklund
 * Copyright (c) 1996, 1997 DiMaga Studios & Innerloop Technologies
 * All rights reserved.
 *
 * This code was originally written for DiMaga Studios by Eivind
 * Eklund.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Neither the name of Eivind Eklund nor the names of DiMaga
 *    Studios or Innerloop Technologies may be used to endorse or
 *    promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#ifndef NODE_H
#define NODE_H
/*lint -esym(740,List_Verify,List_AddTail,List_AddHead,List_RemHead) */
/*lint -esym(740,List_AddTailList,List_AddHeadList,Node_Remove)      */
/*lint -esym(740,Node_InsertBefore,Node_InsertAfter,Node_GetList)    */

/*lint -save +flb */

#define NODE_HEAD(nodetype) /*lint -save -e601 -e19*/ nodetype *pSucc,*pPrev /*lint -restore */
#define LIST_HEAD(nodetype) /*lint -save -e601 -e19*/ nodetype *pHead,*MIDDLE,*pTail /*lint -restore */

struct Node {
	NODE_HEAD(struct Node);
}; /*lint !e770 struct defined identically somewhere else */

struct List {
	LIST_HEAD(struct Node);
}; /*lint !e770 struct defined identically somewhere else */

/* Check that a list is usable.  Will check for non-NULL, initialized list,
   parameter without side-effects, and valid head/tail pointers. */
#define LIST_ASSERT(list) \
  /*lint -save -e506 -e740 -e717 */ \
	do { \
		assert((list) == (list)); /* Check for side-effects */ \
		assert(list); /* Non-NULL */ \
		assert((list)->pHead && (list)->pTail && !(list)->MIDDLE); /* Is initialized */\
		assert((struct Node*)(list)->pHead->pPrev == (struct Node*)(list) && \
			   (struct Node*)(list)->pTail->pSucc == (struct Node*)&(list)->MIDDLE); \
}	while(0) /*lint -restore */

/* Check that a node is usable.  Will check for lack of side-effects,
   non-NULL, and that the node either is a member of a list, or it isn't.
   This last check will catch attempts to use the head or tail node of
   a list. */
		 
#define NODE_ASSERT(node) \
  /*lint -save -e717 -e514 -e506 */ \
	do { \
		assert((node) == (node)); /* Check for side-effects */ \
		assert(node);  /* Non-NULL */ \
		assert(!(!(node)->pSucc ^ !(node)->pPrev)); /* Either both set or both NULL */ \
}	while(0) /*lint -restore */


#if defined(TARGET_PSX)
#  define List_isEmpty(list) /*lint -save -e506 */ ((list)->pHead == (void*)&(list)->MIDDLE) /*lint -restore */
#else
#  define List_isEmpty(list) /*lint -save -e506 */ (assert(list),assert((list)==(list)),(list)->pHead == (void*)&(list)->MIDDLE) /*lint -restore */
#endif

#define List_Init(list) /*lint -save -e717 -e506 */	\
do {																	\
	assert(list);													\
	assert((list)==(list));										\
	(list)->pHead = (void*)&(list)->MIDDLE;				\
	(list)->pTail=(void*)(list);								\
	(list)->MIDDLE=NULL;											\
} while(0) /*lint -restore */

#define Node_Init(node) /*lint -save -e717 -e506 */	\
do {																	\
	assert(node);													\
	assert((node)==(node));										\
	(node)->pSucc=(node)->pPrev=NULL;						\
} while(0) /*lint -restore */

#define Node_isLast(node) ((node)->pSucc->pSucc == NULL)


#define List_AddOrderedOnKey(list,node,type,element)						\
do {																						\
	type* pAOOKNode1;																	\
	type* pAOOKNode2;																	\
	LIST_ASSERT(list);																\
	NODE_ASSERT(node);																\
	for(pAOOKNode1 = (list)->pHead;												\
		 /*lint -save -e720 */														\
		 (pAOOKNode2 = pAOOKNode1->pSucc)										\
			 && pAOOKNode1->element<(node)->element;							\
		 /*lint -restore */															\
		 pAOOKNode1=pAOOKNode2)														\
		;																					\
	/* Cannot use Node_InsertBefore due to possibility of list end */	\
	pAOOKNode1->pPrev->pSucc = (node);											\
	(node)->pPrev = pAOOKNode1->pPrev;											\
	pAOOKNode1->pPrev = (node);													\
	(node)->pSucc = pAOOKNode1;													\
}	/*lint -save -e717 */ while(0) /*lint -restore */


/* Usage:
	struct Node* pNode, pNextNode;
	LIST_WALK_FORWARD(list, pNode, pNextNode) {
		// Do things with pNode - you can even delete it
      // To process one more node, you can do the following:
      if( LIST_WALK_FORWARD_NEXT(pNode, pNextNode) ) {
        // Handle node in pNode
		} else {
        // Handle end of list - MUST end with a break
		  break;
      }
	}
*/

#define LIST_WALK_FORWARD(list,np1,np2) for((np1)=(list)->pHead; /*lint -save -e720 */ ((np2)=(np1)->pSucc) != 0/*lint -restore */; (np1)=(np2))
#define LIST_WALK_BACKWARD(list,np1,np2) for((np1)=(list)->pTail; /*lint -save -e720 */ ((np2)=(np1)->pPrev) != 0 /*lint -restore */; (np1)=(np2))
#define LIST_WALK_FORWARD_NEXT(np1,np2) ((np1)=(np2),((np2)=(np1)->pSucc) != 0)
#define LIST_WALK_BACKWARD_NEXT(np1,np2) ((np1)=(np2),((np2)=(np1)->pPrev) != 0)

const char* List_Verify(const struct List* pList);
void  List_AddTail(struct List* pList, struct Node* pNode);
void  List_AddHead(struct List* pList, struct Node* pNode);
struct Node* List_RemHead(struct List* pList);
void  List_AddTailList(struct List* pList, struct List* pNew);
void  List_AddHeadList(struct List* pList, struct List* pNew);
void  Node_Remove(struct Node* pNode);
void  Node_InsertBefore(struct Node* pNode, struct Node* pNew);
void  Node_InsertAfter(struct Node* pNode, struct Node* pNew);
struct List* Node_GetList(struct Node* pNode);
void  Node_SwapOut(struct Node* pNode, struct Node** ppExternal);
unsigned int  List_Length(const struct List* pList);

/* Make the functions possible to use on any pointer.  WARNING: This
	 disable typechecking.  This will be fixed shortly, but is presently
	 a danger. */

#define List_Verify(pList) (List_Verify(/*lint -save -e740 */(struct List*)(pList)/*lint -restore */))
#define List_AddTail(pList,pNode) (List_AddTail(/*lint -save -e740 */(struct List*)(pList),(struct Node*)(pNode)/*lint -restore */))
#define List_AddHead(pList,pNode) (List_AddHead(/*lint -save -e740 */(struct List*)(pList),(struct Node*)(pNode)/*lint -restore */))
#define List_RemHead(pList) (List_RemHead(/*lint -save -e740 */(struct List*)(pList)/*lint -restore */))
#define List_AddTailList(pList,pNew) (List_AddTailList(/*lint -save -e740 */(struct List*)(pList),(struct List*)(pNew)/*lint -restore */))
#define List_AddHeadList(pList,pNew) (List_AddHeadList(/*lint -save -e740 */(struct List*)(pList),(struct List*)(pNew)/*lint -restore */))
#define Node_Remove(pNode) (Node_Remove(/*lint -save -e740 */(struct Node*)(pNode)/*lint -restore */))
#define Node_InsertBefore(pNode,pNew) (Node_InsertBefore(/*lint -save -e740 */(struct Node*)(pNode),(struct Node*)(pNew)/*lint -restore */))
#define Node_InsertAfter(pNode,pNew) (Node_InsertAfter(/*lint -save -e740 */(struct Node*)(pNode),(struct Node*)(pNew)/*lint -restore */))
#define Node_GetList(pNode) (Node_GetList(/*lint -save -e740 */(struct Node*)(pNode)/*lint -restore */))
#define Node_SwapOut(pNode, ppExternal) (Node_SwapOut(/*lint -save -e740 */(struct Node*)(pNode), (struct Node**)(ppExternal)/*lint -restore */))
#define List_Length(pList) (List_Length(/*lint -save -e740 */(struct List*)(pList)/*lint -restore */))

/* This function is descriped in README.tech */
#define List_SortOnKey(list, type, key)
do {
	 type *data;
	 type *LoC = NULL;			  /* List-of-Chains */
	 type *head;
	 type *tail;
	 type **traverser;
	 type **traverser2;

	 if (List_isEmpty(list))
		 break;
	 if ((list)->pHead == (list)->pTail)
		  break;

	 /* Terminate list, to get to the format we'll be using in this
       function */
	 (list)->pTail->pNext = NULL;
	 /* Split list into a set chains, each sorted (and single-linked). */
	 data = (list)->pHead;
	 while (data) {
		  head = data;
		  data = data->pNext;
		  if (data) {
				tail = head;
				while (data) {
					 if (data->key <= head->key) {
						  data->pPrev = head;
						  head = data;
						  data = data->pNext;
					 } else if (data->key >= tail->key) {
						  tail->pPrev = data;
						  tail = data;
						  data = data->pNext;
					 } else
						  break;
				}
				head->pNext = LoC;
				LoC = head;
				head->pPrev->pNext = tail;
				head = NULL;
		  }
	 }
	 /* Handle singleton if necessary */
	 if (head) {
		  traverser = &LoC;
		  while (*traverser && (*traverser)->key < head->key)
				traverser = &traverser->pPrev;
		  if (traverser = &LoC) {
				/*  Replace head of first chain */
				head->pPrev = LoC;
				head->pNext = LoC->pNext;
				head->pPrev->pNext = LoC->pPrev->pNext;
				LoC = head;
		  } else if (*traverser != NULL) {				
				/*  Normal, inline key in first chain */
				head->pPrev = *traverser;
				*traverser = head;
		  } else {
				/* Last key in first chain */
				*traverser = head;
				LoC->pPrev->pNext = head;
				head->pPrev = NULL;
		  }
	 }
	 /* Merge until there is only a single chain left */
	 while (LoC->pNext) {
		  type **tmp1, **tmp2;
		  type *tmp = NULL;

		  traverser = &LoC;
		  traverser2 = &LoC->pNext;
		  tmp1 = *traverser;
		  tmp2 = *traverser2;
		  /* Remove the second chain - these will be merged */
		  *traverser = (*traverser2)->pNext;
		  for (;;) {
				while (tmp1 && tmp2) {
					 if (tmp1->key < tmp2->key) {
						  tmp->p
					 }
				}
		  }
	 }
} while(0)

/*lint -restore */
#endif /* NODE_H */
