/* -*- mode: c; tab-width: 3; c-basic-indent: 3; -*-
 * Copyright (c) 1998 Eivind Eklund
 * Copyright (c) 1996, 1997 DiMaga Studios & Innerloop Technologies
 * All rights reserved.
 *
 * This code was originally written for DiMaga Studios by Eivind
 * Eklund.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Neither the name of Eivind Eklund nor the names of DiMaga
 *    Studios or Innerloop Technologies may be used to endorse or
 *    promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#include "node.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#define DOTEST(cond,text) /*lint -save -e717 -e506 */do{if(!(cond)) {printf text; printf("\n"); return 0;}} while(0)/*lint -restore*/

/* unit test for the DiMaga list library. */
/* This function test all other functions in the DiMaga list library, */
/* and return failure if they do not work.  It should be updated */
/* each time a new function is added to the library. */
int															/* Did the unit test go well? */
list_Test(void) {
	struct List  TestList;
	struct List  TestList2;
	struct Node  TestNode;
	struct Node  TestNode2;
	struct Node  TestNode3;
	struct Node  TestNode4;
	struct Node* pNode;

	List_Init(&TestList);
	LIST_ASSERT(&TestList);
	DOTEST(!List_Verify(&TestList),("Init - %s\n",List_Verify(&TestList)));
	List_Init(&TestList);

	Node_Init(&TestNode);
	Node_Init(&TestNode2);
	Node_Init(&TestNode3);
	Node_Init(&TestNode4);
	NODE_ASSERT(&TestNode);			/* The rest should be equal */
	List_Init(&TestList2);

	List_AddTail(&TestList,&TestNode);
	DOTEST(!List_Verify(&TestList),("AddTail - %s\n",List_Verify(&TestList)));
	DOTEST(TestList.pHead == &TestNode, ("AddTail - wrong head node."));
	DOTEST(TestList.pTail == &TestNode, ("AddTail - wrong tail node."));
	List_AddTail(&TestList,&TestNode2);
	DOTEST(!List_Verify(&TestList),("AddTail 2 - %s\n",List_Verify(&TestList)));
	DOTEST(TestList.pHead == &TestNode, ("AddTail 2 - wrong head node."));
	DOTEST(TestList.pTail == &TestNode2, ("AddTail 2 - wrong tail node."));

	pNode = List_RemHead(&TestList);
	DOTEST(!List_Verify(&TestList),("RemHead - %s\n",List_Verify(&TestList)));
	DOTEST(pNode,("RemHead - NULL return from non-empty list."));
	DOTEST(pNode == &TestNode, ("RemHead - return of wrong node from non-empty list."));
	pNode = List_RemHead(&TestList);
	DOTEST(!List_Verify(&TestList),("RemHead 2 - %s\n",List_Verify(&TestList)));
	DOTEST(List_isEmpty(&TestList), ("RemHead - List is not empty after removing the only node."));
	DOTEST(!List_RemHead(&TestList),("RemHead - return of node from empty list."));
	DOTEST(!List_Verify(&TestList),("RemHead 3 - %s\n",List_Verify(&TestList)));

	List_AddHead(&TestList,&TestNode);
	DOTEST(!List_Verify(&TestList),("AddHead - %s\n",List_Verify(&TestList)));
	DOTEST(TestList.pHead == &TestNode, ("AddHead - wrong head node."));
	DOTEST(TestList.pTail == &TestNode, ("AddHead - wrong tail node."));
	List_AddHead(&TestList,&TestNode2);
	DOTEST(!List_Verify(&TestList),("AddHead 2 - %s\n",List_Verify(&TestList)));
	DOTEST(TestList.pHead == &TestNode2, ("AddHead 2 - wrong head node."));
	DOTEST(TestList.pTail == &TestNode,  ("AddHead 2 - wrong tail node."));

	List_AddTail(&TestList2,&TestNode3);
	List_AddTail(&TestList2,&TestNode4);
	List_AddTailList(&TestList,&TestList2);
	DOTEST(List_isEmpty(&TestList2),("AddTailList - Did not clear 2nd list."));
	DOTEST(!List_isEmpty(&TestList),("AddTailList - Cleared 1st list"));
	DOTEST(TestList.pHead == &TestNode2,("AddTailList - wrong node at start of target list."));
	DOTEST(TestList.pTail == &TestNode4,("AddTailList - wrong node at end of target list"));
	DOTEST(!List_Verify(&TestList),("AddTailList (target) - %s\n",List_Verify(&TestList)));
	DOTEST(!List_Verify(&TestList2),("AddTailList (source) - %s\n",List_Verify(&TestList2)));

	Node_Remove(&TestNode3);
	DOTEST(!List_Verify(&TestList),("Node_Remove - %s\n",List_Verify(&TestList)));
	Node_Remove(&TestNode4);
	DOTEST(!List_Verify(&TestList2),("Node_Remove (other list) - %s\n",List_Verify(&TestList)));

	List_AddTail(&TestList2,&TestNode3);
	List_AddTail(&TestList2,&TestNode4);
	DOTEST(!List_Verify(&TestList2),("*BEFORE* AddHeadList - %s\n",List_Verify(&TestList2)));
	List_AddHeadList(&TestList,&TestList2);
	DOTEST(!List_Verify(&TestList),("AddHeadList (target) - %s\n",List_Verify(&TestList)));
	DOTEST(!List_Verify(&TestList2),("AddHeadList (source) - %s\n",List_Verify(&TestList2)));
	DOTEST(List_isEmpty(&TestList2),("AddHeadList - Did not clear 2nd list."));
	DOTEST(!List_isEmpty(&TestList),("AddHeadList - Cleared 1st list"));
	DOTEST(TestList.pHead == &TestNode3,("AddHeadList - wrong node at start of target list."));
	DOTEST(TestList.pTail == &TestNode,("AddHeadList - wrong node at end of target list"));

	Node_Remove(&TestNode);
	Node_Remove(&TestNode2);
	Node_Remove(&TestNode3);
	Node_Remove(&TestNode4);
	List_AddTail(&TestList,&TestNode);
	List_AddTail(&TestList,&TestNode2);
	Node_InsertBefore(&TestNode2,&TestNode3);
	DOTEST(!List_Verify(&TestList),("Node_InsertBefore - %s\n",List_Verify(&TestList)));
	DOTEST(TestList.pHead == &TestNode,("Node_InsertBefore - wrong head."));
	DOTEST(TestList.pHead->pSucc == &TestNode3,("Node_InsertBefore - wrong middle."));
	DOTEST(TestList.pHead->pSucc->pSucc == &TestNode2,("Node_InsertBefore - wrong third node."));
	DOTEST(TestList.pTail == &TestNode2,("Node_InsertBefore - wrong tail (too many nodes)."));
	
	DOTEST(Node_GetList(TestList.pHead) == &TestList, ("Node_GetList - wrong list returned for head node."));
	DOTEST(Node_GetList(&TestNode2) == &TestList, ("Node_GetList - wrong list returned for a tail node."));
	DOTEST(!List_Verify(&TestList),("Node_GetList - %s\n",List_Verify(&TestList)));
	Node_InsertAfter(&TestNode2, &TestNode4);
	DOTEST(!List_Verify(&TestList),("Node_InsertAfter - %s\n",List_Verify(&TestList)));
	DOTEST(TestList.pTail == &TestNode4,("Node_InsertAfter insert at the wrong place"));
	
	/* Check that the sorted output is OK */
	{
		typedef struct TestSortedNode {NODE_HEAD(struct TestSortedNode); int vPri;} TSN_t;
		struct TSNList {LIST_HEAD(TSN_t);} sTSNList;
		TSN_t TSN1,TSN2,TSN3,TSN4;

		TSN1.vPri = 1;
		TSN2.vPri = 2;
		TSN3.vPri = 3;
		TSN4.vPri = 4;
		Node_Init(&TSN1);
		Node_Init(&TSN2);
		Node_Init(&TSN3);
		Node_Init(&TSN4);
		List_Init(&sTSNList);
		List_AddOrderedOnKey(&sTSNList,&TSN1,TSN_t,vPri);
		DOTEST(!List_Verify(&TestList),("List_AddOrderedOnKey:1 - %s\n",List_Verify(&TestList)));
		List_AddOrderedOnKey(&sTSNList,&TSN4,TSN_t,vPri);
		List_AddOrderedOnKey(&sTSNList,&TSN2,TSN_t,vPri);
		List_AddOrderedOnKey(&sTSNList,&TSN3,TSN_t,vPri);
		DOTEST(!List_Verify(&TestList),("List_AddOrderedOnKey:2 - %s\n",List_Verify(&TestList)));
		DOTEST(sTSNList.pHead                      == &TSN1,("List_AddOrderedOnKey - wrong #1."));
		DOTEST(sTSNList.pHead->pSucc               == &TSN2,("List_AddOrderedOnKey - wrong #2."));
		DOTEST(sTSNList.pHead->pSucc->pSucc        == &TSN3,("List_AddOrderedOnKey - wrong #3."));
		DOTEST(sTSNList.pHead->pSucc->pSucc->pSucc == &TSN4,("List_AddOrderedOnKey - wrong #4."));
		DOTEST(sTSNList.pTail == &TSN4,("List_AddOrderedOnKey - wrong #4 (tail)."));
	}

	/* Check the Node_SwapOut function */
	{
		List_Init(&TestList);
		Node_Init(&TestNode);
		Node_Init(&TestNode2);
		pNode = &TestNode2;
		List_AddTail(&TestList, &TestNode);

		Node_SwapOut(&TestNode, &pNode);
		DOTEST(!List_Verify(&TestList),("List_SwapOut:1 - %s\n",List_Verify(&TestList)));
		DOTEST(TestList.pHead == &TestNode2,("List_SwapOut - wrong head."));
		DOTEST(TestList.pTail == &TestNode2,("List_SwapOut - wrong tail."));
		DOTEST(pNode          == &TestNode, ("List_SwapOut - wrong pNode."));
	
		Node_SwapOut(&TestNode2, &pNode);
		DOTEST(!List_Verify(&TestList),     ("List_SwapOut:2 - %s\n",List_Verify(&TestList)));
		DOTEST(TestList.pHead == &TestNode, ("List_SwapOut:2 - wrong head."));
		DOTEST(TestList.pTail == &TestNode, ("List_SwapOut:2 - wrong tail."));
		DOTEST(pNode          == &TestNode2,("List_SwapOut:2 - wrong pNode."));
	
	}

	return 1;
}

#include "heap.h"
#include <limits.h>

int
Heap_Test(void) {
	struct HeapTest {
		int TestKey;								/* Sorting key during test */
	} MyHeap[256];
	int nEntries=0;
	struct HeapTest TmpItem;
	size_t HeapSumAdd;						/* Sum of elements in the heap upon add */
	size_t HeapSum;								/* Sum of elements in the normal heap
																	 on remove */
	size_t RevHeapSum;						/* Som of elements in the reverse heap
																	 on remove */
	int LastRemoved;							/* Value of element last removed - for
																	 checking that removals keep order */
	int i;												/* Standard iterator */

	/* Check heap invariant - will assert-fail if it is wrong. */
	Heap_Invariant(MyHeap,&nEntries,TestKey);

	/* Initialize heap with pseudo-random values, keeping the sum of them */
	HeapSumAdd = 0;
	for(i=0; i<10; i++) {
		TmpItem.TestKey = (i*7)^237;
		HeapSumAdd += TmpItem.TestKey;
		MyHeap[i].TestKey = TmpItem.TestKey;
	}

	nEntries = 10;
	Heap_Make(MyHeap,nEntries,TestKey);
	/* Check heap invariant - will assert-fail if it is wrong. */
	Heap_Invariant(MyHeap,&nEntries,TestKey);

	/* Test removal of heap entry, keeping sum */
	LastRemoved = INT_MIN;
	HeapSum = 0;
	while (nEntries) {
		HeapSum += MyHeap[0].TestKey;
		if (MyHeap[0].TestKey < LastRemoved) {
			printf("Elements go off heap in non-sorted order.\n");
			return 0;
		}
		LastRemoved = MyHeap[0].TestKey;
		Heap_Remove(MyHeap,&nEntries,0,TestKey);
		Heap_Invariant(MyHeap,&nEntries,TestKey);
	}
	if (HeapSum != HeapSumAdd) {
		printf("Other elements go off the heap than come on it!\n");
		return 0;
	}

	/* Initialize heap with pseudo-random values */
	HeapSumAdd = 0;
	for(i=0; i<10; i++) {
		TmpItem.TestKey = (i*7)^237;
		HeapSumAdd += TmpItem.TestKey;
		RevHeap_Add(MyHeap,&nEntries,&TmpItem,TestKey);
		/* Check heap invariant - will assert-fail if it is wrong. */
		RevHeap_Invariant(MyHeap,&nEntries,TestKey);
	}

	/* Test removal of heap entry */
	LastRemoved = INT_MAX;
	RevHeapSum = 0;
	while(nEntries) {
		RevHeapSum += MyHeap[0].TestKey;
		if(MyHeap[0].TestKey>LastRemoved) {
			printf("Elements go off reverse heap in non-sorted order.\n");
			return 0;
		}
		LastRemoved = MyHeap[0].TestKey;
		RevHeap_Remove(MyHeap,&nEntries,0,TestKey);
		RevHeap_Invariant(MyHeap,&nEntries,TestKey);
	}
	if( RevHeapSum != HeapSumAdd ) {
		printf("Other elements go off the reverse heap than come on it!\n");
		return 0;
	}
	return 1;
}

#include "./bsearch.h"

int
bsearch_test(void) {
  return 1;
}

int
main(void) {
	return (list_Test() && Heap_Test()) ? EXIT_SUCCESS : EXIT_FAILURE;
}
